<?php

/**
 * Copyright (с) Cloud Linux GmbH & Cloud Linux Software, Inc 2010-2022 All Rights Reserved
 *
 * Licensed under CLOUD LINUX LICENSE AGREEMENT
 * https://www.cloudlinux.com/legal/
 */

namespace XrayProfiler;

if (!class_exists('\XrayProfiler\Log')) {
    class Log
    {
        /**
         * @var array<string>
         *
         * PHP Core Exceptions
         */
        public $core_exceptions = array(
            E_ERROR => 'E_ERROR',                         //1
            E_WARNING => 'E_WARNING',                     //2
            E_PARSE => 'E_PARSE',                         //4
            E_NOTICE => 'E_NOTICE',                       //8
            E_CORE_ERROR => 'E_CORE_ERROR',               //16
            E_CORE_WARNING => 'E_CORE_WARNING',           //32
            E_COMPILE_ERROR => 'E_COMPILE_ERROR',         //64
            E_COMPILE_WARNING => 'E_COMPILE_WARNING',     //128
            E_USER_ERROR => 'E_USER_ERROR',               //256
            E_USER_WARNING => 'E_USER_WARNING',           //512
            E_USER_NOTICE => 'E_USER_NOTICE',             //1024
            E_STRICT => 'E_STRICT',                       //2048
            E_RECOVERABLE_ERROR => 'E_RECOVERABLE_ERROR', //4096
            E_DEPRECATED => 'E_DEPRECATED',               //8192
            E_USER_DEPRECATED => 'E_USER_DEPRECATED',     //16384
            E_ALL => 'E_ALL',                             //32767
        );

        /**
         * @var array<array<int|string|null>>
         */
        private $data = array();

        /**
         * @var self|null
         */
        private static $instance = null;

        private function __construct()
        {
        }

        private function __clone()
        {
        }

        /**
         * @return self
         */
        public static function instance()
        {
            if (is_null(self::$instance)) {
                self::$instance = new self();
                self::$instance->clean();
            }

            return self::$instance;
        }

        /**
         * @return array
         */
        public function getData()
        {
            return $this->data;
        }

        /**
         * @param int $errno
         * @param string $errstr
         * @param ?string $errfile
         * @param ?int $errline
         *
         * @return void
         */

        public function setData($errno, $errstr, $errfile = null, $errline = null)
        {
            $this->data[] = array(
                'message' => $errstr,
                'type' => isset($this->core_exceptions[$errno])
                    ? $this->core_exceptions[$errno]
                    : 'Undefined: ' . $errno,
                'filename' => $errfile,
                'lineno' => $errline,
            );
        }

        /**
         * @return string|bool
         */
        public function sendData()
        {
            $data = $this->prepareSentryData($this->data);
            if (empty($data)) {
                return false;
            }
            $sentry_response = $this->sendSentryData($data);
            if (!$sentry_response) {
                return false;
            }
            $this->clean();
            return $sentry_response;
        }

        /**
         * @param array $data
         *
         * @return string|bool
         */
        private function prepareSentryData($data)
        {
            if (empty($data)) {
                return false;
            }
            $formatted = array();
            foreach ($data as $row) {
                $formatted[] = array(
                    "stacktrace" => array(
                        "frames" => array(
                            array(
                                "filename" => $row["filename"],
                                "lineno" => $row["lineno"],
                            )
                        ),
                        "frames_omitted" => null,
                    ),
                    "type" => $row['type'],
                    "value" => $row['message'],
                );
            }
            $sentry = array(
                "extra" => array(
                    "php_version" => phpversion(),
                    "xray_version" => defined('XRAY_RELEASE') ? XRAY_RELEASE : null,
                ),
                "sentry.interfaces.Exception" => array(
                    "exc_omitted" => null,
                    "values" => $formatted,
                ),
            );
            return json_encode($sentry);
        }

        /**
         * @param string|bool $data
         *
         * @return string|bool
         */
        private function sendSentryData($data)
        {
            if (!function_exists('curl_init') || empty($data)) {
                return false;
            }

            $ch = curl_init();
            $sentry_key = '8e6821f19d214977ace88586bc8569fd';
            $url = 'https://' . $sentry_key . '@cl.sentry.cloudlinux.com/api/23/store/';

            curl_setopt($ch, CURLOPT_HTTPHEADER, array(
                'Content-Type: application/json',
                'X-Sentry-Auth: Sentry sentry_version=7,sentry_timestamp=' . time()
                . ',sentry_client=php-curl/1.0,sentry_key=' . $sentry_key,
            ));
            curl_setopt($ch, CURLOPT_URL, $url);
            curl_setopt($ch, CURLOPT_POST, true);
            curl_setopt($ch, CURLOPT_CUSTOMREQUEST, 'POST');
            curl_setopt($ch, CURLOPT_POSTFIELDS, $data);
            curl_setopt($ch, CURLOPT_TIMEOUT, 1);
            curl_setopt($ch, CURLOPT_CONNECTTIMEOUT, 1);
            curl_setopt($ch, CURLOPT_HEADER, 0);
            curl_setopt($ch, CURLOPT_RETURNTRANSFER, true);
            $response = curl_exec($ch);
            curl_close($ch);

            return $response;
        }

        /**
         * @return void
         */
        public function clean()
        {
            $this->data = array();
        }
    }
}
