<?php
/**
 * Copyright (с) Cloud Linux GmbH & Cloud Linux Software, Inc 2010-2022 All Rights Reserved
 *
 * Licensed under CLOUD LINUX LICENSE AGREEMENT
 * https://www.cloudlinux.com/legal/
 */

namespace CloudLinux\SmartAdvice\App\Advice;

use CloudLinux\SmartAdvice\App\Option;
use CloudLinux\SmartAdvice\App\Notice\Model as NoticeModel;
use Exception;

/**
 * Advice manager
 */
class Manager {
	/**
	 * Api.
	 *
	 * @var Api
	 */
	private $api;

	/**
	 * Option.
	 *
	 * @var Option
	 */
	private $option;

	/**
	 * Site domain.
	 *
	 * @var string
	 */
	private $domain = '';

	/**
	 * WordPress path.
	 *
	 * @var string
	 */
	private $website = '';

	/**
	 * Advices.
	 *
	 * @var array<Model>
	 */
	protected $advices;

	/**
	 * Constructor.
	 *
	 * @param  Api    $api  api.
	 * @param  Option $option  api.
	 */
	public function __construct( Api $api, Option $option ) {
		$this->api     = $api;
		$this->option  = $option;
		$this->advices = $this->all();
		$this->env();

		add_action( 'cl_smart_advice_sync', array( $this, 'sync' ), 10, 1 );
		add_action( 'cl_smart_advice_sync_statuses', array( $this, 'syncStatuses' ), 10, 0 );
		add_action( 'cl_smart_advice_status', array( $this, 'status' ), 10, 2 );
		add_action( 'cl_smart_advice_apply', array( $this, 'apply' ), 10, 2 );
		add_action( 'cl_smart_advice_rollback', array( $this, 'rollback' ), 10, 2 );
	}

	/**
	 * Evn info.
	 *
	 * @return void
	 */
	private function env() {
		if ( defined( 'WP_HOME' ) ) {
			$home = (string) WP_HOME;
		} else {
			$home = (string) get_option( 'home' );
		}

		$parse = wp_parse_url( $home );

		if ( ! is_array( $parse ) ) {
			$parse = array();
		}

		$domain = array_key_exists( 'host', $parse ) ? $parse['host'] : '';
		$path   = array_key_exists( 'path', $parse ) ? $parse['path'] : '';

		$this->domain  = $domain;
		$this->website = ! empty( $path ) ? $path : '/';
	}

	/**
	 * Domain.
	 *
	 * @return string
	 */
	public function domain() {
		return $this->domain;
	}

	/**
	 * Website.
	 *
	 * @return string
	 */
	public function website() {
		return $this->website;
	}

	/**
	 * Api interface.
	 *
	 * @return Api
	 */
	private function api() {
		return $this->api;
	}

	/**
	 * Advices array.
	 *
	 * @return array<Model>
	 */
	public function advices() {
		return $this->advices;
	}

	/**
	 * Check domain and WordPress path.
	 *
	 * @param  array $item  data.
	 *
	 * @return boolean
	 */
	private function checkMetadata( $item ) {
		if ( strtolower( $this->domain() ) !== strtolower( $item['metadata']['domain'] ) ) {
			return false;
		}

		if ( strtolower( $this->website() ) !== strtolower( $item['metadata']['website'] ) ) {
			return false;
		}

		return true;
	}

	/**
	 * Check advice.
	 *
	 * @param  array $advice  advice.
	 *
	 * @return boolean
	 */
	private function checkAdvice( $advice ) {
		if ( 'outdated' === $advice['status'] ) {
			return false;
		}

		return true;
	}

	/**
	 * Add or update advice.
	 *
	 * @param  array $advice  data.
	 *
	 * @return self
	 */
	private function set( $advice ) {
		if ( array_key_exists( $advice['id'], $this->advices() ) ) {
			$model = $this->advices()[ $advice['id'] ];
		} else {
			$model = new Model();
		}

		$model->fill( $advice );

		$this->advices[ $advice['id'] ] = $model;

		return $this;
	}

	/**
	 * Sync advice.
	 *
	 * @param  array $response  advices.
	 *
	 * @return void
	 */
	public function sync( $response ) {
		if ( is_array( $response ) && array_key_exists( 'data', $response ) ) {
			$this->advices = array();
			$items         = array_filter( $response['data'], array( $this, 'checkMetadata' ) );
			foreach ( $items as $item ) {
				$advice = $item['advice'];

				$advice['created_at'] = $item['created_at'];
				$advice['updated_at'] = $item['updated_at'];

				if ( ! $this->checkAdvice( $advice ) ) {
					continue;
				}

				$this->set( $advice );
			}

			$this->save();
		} else {
			add_filter( 'cl_smart_advice_sync_failed', '__return_true' );
		}
	}

	/**
	 * Sync statuses.
	 *
	 * @return void
	 */
	public function syncStatuses() {
		$advices = $this->advices();

		foreach ( $advices as $advice ) {
			$this->status( $advice->id, false );
		}

		$this->save();
	}

	/**
	 * Advice modifier.
	 *
	 * @param  int      $advice_id  advice.
	 * @param  callable $callback  command.
	 * @param  bool     $save  advice.
	 *
	 * @return void
	 */
	private function modifier( $advice_id, $callback, $save = true ) {
		$advices = $this->advices();

		if ( ! array_key_exists( $advice_id, $advices ) ) {
			do_action( 'cl_smart_advice_notice_add', $advice_id, NoticeModel::ERROR_ADVICE_NOT_FOUND['type'], array( 'advice_id' => $advice_id ) );

			return;
		}

		$advice = $advices[ $advice_id ];

		if ( is_callable( $callback ) ) {
			$callback( $advice );
		}

		$this->set( $advice->toArray() );

		if ( true === $save ) {
			$this->save();
		}
	}

	/**
	 * Apply advice, set status pending.
	 *
	 * @param  int  $advice_id  to apply.
	 * @param  bool $save  advices.
	 *
	 * @return void
	 */
	public function apply( $advice_id, $save = true ) {
		$this->modifier(
			$advice_id,
			function ( $advice ) {
				$result = $this->api()->apply( $advice->id );
				if ( ! $result ) {
					do_action( 'cl_smart_advice_set_error', E_USER_WARNING, 'Advice manager: failed modifier apply for ' . $advice->id, __FILE__, __LINE__ );
					do_action( 'cl_smart_advice_notice_add', $advice->id, NoticeModel::ERROR_ADVICE_APPLY['type'], array() );
				} else {
					$advice->status = 'pending';
				}
			},
			$save
		);
	}

	/**
	 * Rollback advice, set status pending.
	 *
	 * @param  int  $advice_id  id.
	 * @param  bool $save  advices.
	 *
	 * @return void
	 */
	public function rollback( $advice_id, $save = true ) {
		$this->modifier(
			$advice_id,
			function ( $advice ) {
				$result = $this->api()->rollback( $advice->id );
				if ( ! $result ) {
					do_action( 'cl_smart_advice_set_error', E_USER_WARNING, 'Advice manager: failed modifier rollback for ' . $advice->id, __FILE__, __LINE__ );
					do_action( 'cl_smart_advice_notice_add', $advice->id, NoticeModel::ERROR_ADVICE_ROLLBACK['type'], array() );
				} else {
					$advice->status = 'pending';
				}
			},
			$save
		);
	}

	/**
	 * Update advice status.
	 *
	 * @param  int  $advice_id  id.
	 * @param  bool $save  advices.
	 *
	 * @return void
	 */
	public function status( $advice_id, $save = true ) {
		$this->modifier(
			$advice_id,
			function ( $advice ) {
				$json = $this->api()->status( $advice->id );

				$response = json_decode( $json, true );
				if ( is_array( $response ) ) {
					if ( array_key_exists( 'status', $response ) ) {
						$advice->status = $response['status'];
					}

					if ( array_key_exists( 'total_stages', $response ) ) {
						$advice->total_stages = $response['total_stages'];
					}

					if ( array_key_exists( 'completed_stages', $response ) ) {
						$advice->completed_stages = $response['completed_stages'];
					}

					if ( array_key_exists( 'data', $response ) && array_key_exists( 'result', $response['data'] ) ) {
						$result = $response['data']['result'];
						if ( 'success' !== $result ) {
							$description = $result;

							if ( array_key_exists( 'context', $response['data'] ) ) {
								$search  = array();
								$replace = array();
								foreach ( $response['data']['context'] as $key => $val ) {
									$search[]  = '%(' . $key . ')s';
									$replace[] = $val;
								}
								$description = str_replace( $search, $replace, $result );
							}

							do_action( 'cl_smart_advice_notice_add', $advice->id, NoticeModel::ERROR_ADVICE_APPLY['type'], array( 'description' => $description ) );
						}
					}
				}

				return $json;
			},
			$save
		);
	}

	/**
	 * All cached advices.
	 *
	 * @return array
	 */
	private function all() {
		$advices = $this->option->get( 'advices' );

		return array_map(
			function ( $advice ) {
				return ( new Model() )->fill( $advice );
			},
			$advices
		);
	}

	/**
	 * Time sync at.
	 *
	 * @return int|null
	 */
	public function syncAt() {
		return $this->option->get( 'sync_at' );
	}

	/**
	 * Update advices' cache.
	 *
	 * @return void
	 */
	public function save() {
		$advices = $this->advices();

		$this->option->save( 'advices', $advices );
		$this->option->save( 'sync_at', time() );
	}

	/**
	 * Has pending advices.
	 *
	 * @return bool
	 */
	public function hasPending() {
		$advices = $this->advices();
		$pending = array_filter(
			$advices,
			function ( $advice ) {
				return 'pending' === $advice->status;
			}
		);

		return count( $pending ) > 0;
	}

	/**
	 * Plugin compatible.
	 *
	 * @return bool
	 */
	public function compatible() {
		return function_exists( 'stream_socket_client' );
	}
}
